@@ -10,10 +10,11 @@
    as published by the Free Software Foundation; either version
    2 of the License, or (at your option) any later version.
 */
-/* $Id: ciped.c,v 1.22.2.3 1999/06/07 19:08:24 olaf Exp $ */
+/* $Id: ciped.c,v 1.34 2000/07/28 01:12:04 olaf Exp $ */
 
 #define CTLDIR "/etc/cipe"
 
+#define _GNU_SOURCE
 #include <ctype.h>
 #include <errno.h>
 #include <fcntl.h>
@@ -21,13 +22,12 @@
 #include <signal.h>
 #include <stdio.h>
 #include <stdlib.h>
-#define __USE_GNU
 #include <string.h>
 #include <syslog.h>
 #include <unistd.h>
 #include <arpa/inet.h>
-#include <net/if.h>
 #include <sys/ioctl.h>
+#include <sys/mman.h>
 #include <sys/socket.h>
 #include <sys/stat.h>
 #include <sys/time.h>
@@ -44,86 +44,13 @@ static const char daemon_version[]=VERSION;
 
 void usage(const char *p)
 {
-    fprintf(stderr, "usage: %s [-o file ] [options] device=name\n", p);
+    fprintf(stderr, "usage: %s [i] [-o file ] [option...]\n", p);
     exit(1);
 }
 
 /* FIXME: places marked *#* have byteorder conversion missing */
 /* This is corrected in version 3 protocol */
 
-/*** The global options structure ***/
-
-struct options {
-    const char * const oname;
-    const char otyp;
-    union {
-	int ovint;
-	char *ovstr;
-	struct sockaddr_in *ovsaddr;
-    } ov;
-} opts[] = {
-#define Opt(x,t) { #x, T##t, {0} },
-#define Tbool	0
-#define Tint	1
-#define Tstr	2
-#define Taddr	3
-#define Tuaddr	4
-#define Ttaddr	5
-
-#define Okey		0  /* The interface secret key
-			      - set via options file only */
-Opt(key,str)
-#define Odebug		1  /* Debug mode - don't detach, verbose messages */
-Opt(debug,bool)
-#define Oipaddr		2  /* This CIPE device's IP address */
-Opt(ipaddr,addr)
-#define Optpaddr	3  /* The peer CIPE device's IP address */
-Opt(ptpaddr,addr)
-#define Omtu		4  /* Device MTU */
-Opt(mtu,int)
-#define Ometric		5  /* Device metric */
-Opt(metric,int)
-#define Ome		6  /* My UDP address */
-Opt(me,uaddr)
-#define Opeer		7  /* The peer's UDP address,
-			      i.e. where to send packets */
-Opt(peer,uaddr)
-#define Odevice		8  /* name of the CIPE device, e.g. cip3b0 */
-Opt(device,str)
-#define Onokey		9  /* Don't encrypt */
-Opt(nokey,bool)
-#define Osocks		10  /* SOCKS5 server (ip:port) */
-Opt(socks,taddr)
-#define Otokxc		11  /* Timeout for key exchange (secs) */
-Opt(tokxc,int)
-#define Otokey		12  /* Dynamic key lifetime (secs) */
-Opt(tokey,int)
-#define Oipup		13  /* Name of the ip-up script */
-Opt(ipup,str)
-#define Oipdown		14  /* Name of the ip-down script */
-Opt(ipdown,str)
-#define Oarg		15  /* Extra argument to ip-up script */
-Opt(arg,str)
-#define Octtl		16  /* TTL value for carrier */
-Opt(cttl,int)
-#define Omaxerr		17  /* Maximum ECONNREFUSED tolerance */
-Opt(maxerr,int)
-#define Otokxts         18  /* Timeout for key exchange timestamps */
-Opt(tokxts,int)
-#define Oping           19  /* Keepalive ping interval */
-Opt(ping,int)
-#define Otoping         20  /* Keepalive ping timeout */
-Opt(toping,int)
-
-#undef Opt
-{ NULL, 0, {0} }};
-
-#define OI(x)      (opts[O##x].ov.ovint)
-#define OS(x)      (opts[O##x].ov.ovstr)
-#define OA(x)      (opts[O##x].ov.ovsaddr)
-#define OAaddr(x)  (OA(x)->sin_addr)
-#define OAport(x)  (OA(x)->sin_port)
-
 /* Statistics */
 struct kstat {
     int rreq, req, ind, indb, ack, ackb, bogus;
@@ -131,6 +58,97 @@ struct kstat {
 
 char device[16] = "";
 int devnum = -1;
+int inetd = 0;
+
+#ifdef VER_CRC32
+typedef unsigned long crc;
+#define CRCFORM "%08lX"
+#else
+typedef unsigned short crc;
+#define CRCFORM "%04X"
+#endif
+
+/*** Locked memory structure, contains everything to do with keys */
+#define SECPOOL 256
+struct lockmem {
+    char pool[SECPOOL];                         /* for strings */
+    int poolptr;
+    struct siocsifcipkey sio;                   /* setkey ioctl */
+    unsigned char kxbuf[KEYXCHGBLKMIN+2];       /* KX comm buff */
+    crc scrc;
+    unsigned char skey[userKeySize];            /* holds current KX key */
+} *LM;
+
+char *secstrdup(const char *p)
+{
+    char *q=LM->pool+LM->poolptr;
+    int n=strlen(p)+1;
+    if (LM->poolptr+n>SECPOOL)
+        return NULL;
+    strcpy(q, p);
+    LM->poolptr+=n;
+    return q;
+}
+
+/* Check file and path up to it for wrong permissions. Note: this is
+   only a first-order admin stupidity test, not attack-proofing. */
+int secchk(const char *f, int fmask, int dmask, int v)
+{
+    struct stat s;
+    char *b=strdup(f), *p;
+
+    if (!b) {
+        fprintf(stderr, "strdup error\n");
+        return -1;
+    }
+#define err(f) do{ logs(LOG_ERR, f, b); goto error; }while(0)
+
+    if (b[0]!='/')
+        err("%s: not absolute");
+    if (stat(b, &s)<0) {
+        if (v)
+            logs(LOG_ERR, "%s: stat: %m", b);
+        goto error;
+    }
+    if (!S_ISREG(s.st_mode) || (s.st_mode&fmask) || (s.st_uid!=0))
+        err("%s: incorrect permissions");
+    while ((p=strrchr(b, '/'))) {
+        if (p==b)
+            p[1]='\0';
+        else
+            *p='\0';
+        if (stat(b, &s)<0)
+            err("%s: stat: %m");
+        if (!S_ISDIR(s.st_mode) || (s.st_mode&dmask) || (s.st_uid!=0))
+            err("%s: incorrect permissions");
+        if (p==b)
+            break;
+    }
+    free(b);
+    return 0;
+
+ error:
+    free(b);
+    return -1;
+#undef err
+}
+
+#ifdef HAVE_MLOCK
+void xmlock(void *addr, size_t len)
+{
+    if (mlock(addr, len)<0)
+        logerr(LOG_WARNING, "mlock"); /* not fatal */
+}
+void xmunlock(void *addr, size_t len)
+{
+    if (munlock(addr, len)<0)
+        logerr(LOG_WARNING, "munlock"); /* not fatal */
+}
+#else
+#define xmlock(a,l)
+#define xmunlock(a,l)
+#endif
+
 
 /*** Option parser ***/
 
@@ -172,7 +190,7 @@ int getaddr(char *c, struct sockaddr_in *sa, const char *prot)
 }
 
 /* Set an option */
-int setopt(const char *n, char *v)
+int setopt(const char *n, char *v, int r)
 {
     struct options *o=opts;
     static const char * const protos[]={NULL, "udp", "tcp"};
@@ -184,9 +202,11 @@ int setopt(const char *n, char *v)
 	    }
 	    switch(o->otyp) {
 	    case Tbool:
-		o->ov.ovint=1; break;
+		o->ov.ovint=1;
+		break;
 	    case Tint:
-		o->ov.ovint=atoi(v); break;
+		o->ov.ovint=atoi(v);
+		break;
 	    case Tstr:
 		if (!(o->ov.ovstr=strdup(v))) {
 		    fprintf(stderr, "setopt: out of memory\n"); return -1;
@@ -202,6 +222,16 @@ int setopt(const char *n, char *v)
 		    return -1;
 		}
 		break;
+            case Tsecret:
+                if (r) {
+                    fprintf(stderr, "Refused to set key from command line!\n");
+                    return -1;
+                }
+		if (!(o->ov.ovstr=secstrdup(v))) {
+		    fprintf(stderr, "setopt: out of secret-pool memory\n");
+                    return -1;
+		}
+		break;
 	    default:
 		fprintf(stderr, "internal error: %d\n", __LINE__); return -1;
 	    }
@@ -221,43 +251,26 @@ void setopt_cmdline(int argc, char *argv[])
     for (i=0; i<argc; ++i) {
 	if ((q=strchr(argv[i], '=')))
 	    *q++='\0';
-	if (!strcmp(argv[i], "key")) {
-	    fprintf(stderr, "Refused to set key from command line!\n");
-	    continue;
-	}
-	(void) setopt(argv[i], q);
+	(void) setopt(argv[i], q, 1);
     }
 }
 
 /* Set options from file, report file not found if (v) */
 void setopt_file(const char *n, int v)
 {
-    int d;
     FILE *f;
     char b[128];
     char *p, *q;
-    struct stat s;
 
-    if ((d=open(n, O_RDONLY))<0) {
+    if (secchk(n, 0077, 0022, v))
+        return;
+    if (!(f=fopen(n, "r"))) {
 	if (v)
 	    perror("setopt_file: open");
 	return;
     }
-    if (fstat(d, &s)<0) {
-	perror("setopt_file: stat");
-	return;
-    }
-    if (!S_ISREG(s.st_mode) || (s.st_mode&(S_IRWXG|S_IRWXO)) ||
-	(s.st_uid!=0)) {
-	fprintf(stderr, "file %s has incorrect permissions\n", n);
-	return;
-    }
-
-    if (!(f=fdopen(d, "r"))) {
-	if (v)
-	    perror("setopt_file: fdopen");
-	return;
-    }
+    xmlock(b, sizeof(b));
+    /* FIXME: perhaps fgets could allocate unlocked memory */
     while (fgets(b, sizeof(b), f)) {
 	p=b;
 	if ((q=strchr(p, '\n')))
@@ -272,8 +285,10 @@ void setopt_file(const char *n, int v)
 	    if (!*q)
 		q=NULL;
 	}
-	(void) setopt(p, q);
+	(void) setopt(p, q, 0);
+        memset(b, 0, sizeof(b));
     }
+    xmunlock(b, sizeof(b));
     fclose(f);
 }
 
@@ -293,6 +308,10 @@ char *optstr(struct options *o)
         snprintf(buf, sizeof(buf), "%s=%s", o->oname,
                  o->ov.ovstr?o->ov.ovstr:"(none)");
         break;
+    case Tsecret:
+        snprintf(buf, sizeof(buf), "%s=%s", o->oname,
+                 o->ov.ovstr?"(secret)":"(none)");
+        break;
     case Taddr:
         if (o->ov.ovsaddr)
             snprintf(buf, sizeof(buf), "%s=%s", o->oname,
@@ -360,20 +379,6 @@ void argmiss(const char *s)
 /* print if debugging */
 #define dprintf(f)    if(OI(debug)){printf f;}
 
-void dumppacket(const char *hdr, int f, const unsigned char *buf, int len)
-{
-    int i=999;
-    printf("%s %d", hdr, f);
-    while (len-->0) {
-        if (++i>15) {
-            printf("\n    ");
-            i=0;
-        }
-        printf("%02x ", *buf++);
-    }
-    printf("\n");
-}
-
 /*** Initialize, open and attach the CIPE device. ***/
 
 int opendev(void)
@@ -383,17 +388,20 @@ int opendev(void)
     struct siocsifcipall d;
 #endif
     struct siocsifcippar p;
-    struct siocsifcipkey k;
     struct siocsifcipatt a;
     struct ifreq r;
 
-#define amiss(s) { argmiss(s); goto error; }
-#define err(s)   { logerr(LOG_ERR, s); goto error; }
+#define amiss(s) do{ argmiss(s); goto error; }while(0)
+#define err(s)   do{ logerr(LOG_ERR, s); goto error; }while(0)
 
     /* Initialize, bind and connect the socket */
-    if ((f=socket(AF_INET, SOCK_DGRAM, 0))<0) {
-	logerr(LOG_ERR, "opendev: socket");
-	return -1;
+    if (inetd) {
+        f=0;
+    } else {
+        if ((f=socket(AF_INET, SOCK_DGRAM, 0))<0) {
+            logerr(LOG_ERR, "opendev: socket");
+            return -1;
+        }
     }
     i=65536; /* we need much to receive fast */
     if (setsockopt(f, SOL_SOCKET, SO_RCVBUF, &i, sizeof(i)))
@@ -406,13 +414,15 @@ int opendev(void)
 	OAaddr(me).s_addr=INADDR_ANY;
 	OAport(me)=0;
     }
-    if (bind(f, (struct sockaddr *)OA(me), sizeof(*OA(me)))<0)
-	err("opendev: bind");
-    if (OA(peer)) {
-	if (connect(f, (struct sockaddr *)OA(peer), sizeof(*OA(peer)))<0)
-	    err("opendev: bind");
-    } else
-	amiss("peer");
+    if (!inetd) {
+        if (bind(f, (struct sockaddr *)OA(me), sizeof(*OA(me)))<0)
+            err("opendev: bind");
+        if (OA(peer)) {
+            if (connect(f, (struct sockaddr *)OA(peer), sizeof(*OA(peer)))<0)
+                err("opendev: bind");
+        } else
+            amiss("peer");
+    }
     i=sizeof(*OA(me));
     if (getsockname(f, (struct sockaddr *)OA(me), &i)<0)
 	/* not fatal */
@@ -421,6 +431,9 @@ int opendev(void)
 #ifdef NO_DYNDEV
     if (!OS(device))
         amiss("device");
+    if (strlen(OS(device)) > 15)
+        err("opendev: device name too long.");
+    strcpy(device, OS(device));
 #else
     /* Allocate the device name */
     if (OS(device)) {
@@ -446,7 +459,11 @@ int opendev(void)
     }
     p.tmo_keyxchg=OI(tokxc);
     p.tmo_keylife=OI(tokey);
-    p.mayclear=OI(nokey);
+    p.flags=CIPF_MAY_STKEY;
+    if (OI(nokey))
+        p.flags|=CIPF_MAY_CLEAR;
+    if (OI(dynip))
+        p.flags|=CIPF_MAY_DYNIP;
     p.cttl=OI(cttl);
     if (ioctl_setpar(f, device, &p)<0)
 	err("opendev: setpar");
@@ -454,10 +471,11 @@ int opendev(void)
     /* Set the key */
     if (OS(key)) {
 	unsigned char *c=OS(key);
-	unsigned char *x=(unsigned char *)&k.thekey;
+	unsigned char *x=(unsigned char *)&LM->sio.thekey;
 	i=0;
-	k.which=KEY_STATIC;
-	/* primitive hex decoder */
+	LM->sio.which=KEY_STATIC;
+#ifdef BUG_COMPATIBLE
+	/* BROKEN attempt at a primitive hex decoder */
 	while(*c) {
 	    x[i]=((*c)<<4)&0xF0;
 	    *c++='\0'; /* clear after use */
@@ -466,9 +484,21 @@ int opendev(void)
 	    *c++='\0';
 	    if (i>=userKeySize) break;
 	}
-	if (ioctl_setkey(f, device, &k)<0)
+#else
+	/* simple hex decoder */
+	#define dehex(x) (((x)>='0'&&(x)<='9')?((x)-'0'):(((x)&15)+9))
+	while (*c) {
+	    x[i]=dehex(*c)<<4;
+	    *c++='\0'; /* clear after use */
+	    if (!*c) break;
+	    x[i++]|=dehex(*c);
+	    *c++='\0';
+	    if (i>=userKeySize) break;
+	}
+#endif
+	if (ioctl_setkey(f, device, &LM->sio)<0)
 	    err("opendev: setkey");
-	memset(&k, 0, sizeof(k));
+	memset(&LM->sio, 0, sizeof(struct siocsifcipkey));
     } else {
 	if (!OI(nokey))
 	    amiss("key");
@@ -537,7 +567,7 @@ void closedev(int f)
         }
         x.num=devnum;
         if (ioctl_unalloc(f, device, &x)<0)
-            logerr(LOG_ERR, "closedev: alloc");
+            logerr(LOG_ERR, "closedev: unalloc");
         close(f);
     }
 #endif
@@ -549,7 +579,7 @@ char *getstats(void)
 {
     static char b[256];
     char *p, *q;
-    FILE *f=fopen("/proc/net/dev/", "r");
+    FILE *f=fopen("/proc/net/dev", "r");
     if (!f)
         return NULL;
     while (fgets(b, sizeof(b), f)) {
@@ -621,7 +651,7 @@ void sighand(int s)
     gotsig=s;
 }
 
-void sigignore(int s)
+void sighandnot(int s)
 {
     /* Dummy */
 }
@@ -632,6 +662,10 @@ void setsig(int sig, void (*fun)(int))
   sa.sa_handler = fun;
   sa.sa_flags = 0;
   sigemptyset(&sa.sa_mask);
+  sigaddset(&sa.sa_mask, SIGHUP);
+  sigaddset(&sa.sa_mask, SIGINT);
+  sigaddset(&sa.sa_mask, SIGTERM);
+  sigaddset(&sa.sa_mask, SIGCHLD);
   sigaction(sig, &sa, NULL);
 }
 
@@ -652,14 +686,16 @@ void dprintct(const char *h, char c)
 }
 
 /* send a control */
-int sendctl(int f, const char *msg, int len)
+int sendctl(int f, const char *msg, int len, int v)
 {
     int e;
     dprintct("sending", msg[0]);
     if (OI(nokey))
         return -1; /* unencrypted cant handle controls */
-    if ((e=send(f, msg, len, 0))<0)
-        logerr(LOG_ERR, "sendctl");
+    if ((e=send(f, msg, len, 0))<0) {
+	if (v)
+	    logerr(LOG_ERR, "sendctl");
+    }
     return e;
 }
 
@@ -679,7 +715,7 @@ void ctrl(int f, char *msg, int len)
         break;
     case CT_PING:
         msg[0]=CT_PONG;
-        sendctl(f, msg, len);
+        sendctl(f, msg, len, 1);
         break;
     case CT_PONG:
         pongtime=MAXINT;
@@ -695,14 +731,6 @@ void ctrl(int f, char *msg, int len)
 
 /*** The key exchange procedure ***/
 
-#ifdef VER_CRC32
-typedef unsigned long crc;
-#define CRCFORM "%08lX"
-#else
-typedef unsigned short crc;
-#define CRCFORM "%04X"
-#endif
-
 #define REM 	256	/* How many keys to remember */
 
 /* Cache last valid key CRCs. */
@@ -731,32 +759,24 @@ int errcnt;
 int kxchg(int f, int r)
 {
     int x=0, e=0;
-    unsigned char kx[KEYXCHGBLKMIN+2];
-    #define kx_typ       (kx[0])
-    #define kx_nkind_key (kx+1)
-    #define kx_ts        (*(time_t *)(kx+KEYXCHGTSPOS))
+    crc c;
+
+    #define kx_typ       (LM->kxbuf[0])
+    #define kx_nkind_key (LM->kxbuf+1)
+    #define kx_ts        (*(time_t *)(LM->kxbuf+KEYXCHGTSPOS))
 #ifdef VER_CRC32
-    #define kx_nkind_crc (*(unsigned long *)(kx+1+userKeySize))
-    #define kx_nkack_crc (*(unsigned long *)(kx+1))
+    #define get_kx_nkind_crc(x) memcpy((x), LM->kxbuf+1+userKeySize, 4)
+    #define put_kx_nkind_crc(x) memcpy(LM->kxbuf+1+userKeySize, (x), 4)
+    #define get_kx_nkack_crc(x) memcpy((x), LM->kxbuf+1, 4)
+    #define put_kx_nkack_crc(x) memcpy(LM->kxbuf+1, (x), 4)
 #else
-    #define kx_nkind_crc (*(unsigned short *)(kx+1+userKeySize))
-    #define kx_nkack_crc (*(unsigned short *)(kx+1))
+    #define get_kx_nkind_crc(x) memcpy((x), LM->kxbuf+1+userKeySize, 2)
+    #define put_kx_nkind_crc(x) memcpy(LM->kxbuf+1+userKeySize, (x), 2)
+    #define get_kx_nkack_crc(x) memcpy((x), LM->kxbuf+1, 2)
+    #define put_kx_nkack_crc(x) memcpy(LM->kxbuf+1, (x), 2)
 #endif
-    static crc scrc=0;
-    static unsigned char skey[userKeySize];
 
-/*
-    struct msghdr hdr;
-    struct iovec iov;
-
-    iov.iov_base=&kx;
-    iov.iov_len=sizeof(kx)-1;
-    hdr.msg_name=hdr.msg_control=NULL;
-    hdr.msg_iov=&iov;
-    hdr.msg_iovlen=1;
-    hdr.msg_flags=0;
-*/
-    if ((x=recv(f, kx, KEYXCHGBLKMIN, 0))<=0) {
+    if ((x=recv(f, LM->kxbuf, KEYXCHGBLKMIN, 0))<=0) {
 	if (errno==EAGAIN || errno==EINTR)
 	     /* This is normal, would be even after select */
 	    return 0;
@@ -768,10 +788,9 @@ int kxchg(int f, int r)
 	return (OI(maxerr)<0 ? 0 : --errcnt);
     }
     errcnt=OI(maxerr);
-    /* dumppacket("recvmsg", hdr.msg_flags, kx, x); */
-    if (kx[0]>=CT_DUMMY) {
-        kx[55]='\0'; /* max length of message */
-        ctrl(f, kx, x);
+    if (LM->kxbuf[0]>=CT_DUMMY) {
+        LM->kxbuf[55]='\0'; /* max length of message */
+        ctrl(f, LM->kxbuf, x);
         return 0;
     }
     if (OI(tokxts) && (abs(time(0)-kx_ts)>OI(tokxts))) {
@@ -782,23 +801,24 @@ int kxchg(int f, int r)
     case NK_RREQ:
 	kx_typ=NK_REQ;
 	dprintf(("KX: [NK_RREQ] sending NK_REQ\n"));
-	x=send(f, &kx, sizeof(kx), 0); e=errno;
+	x=send(f, &LM->kxbuf, sizeof(LM->kxbuf), 0); e=errno;
 	++ks.rreq;
 	break;
     case NK_REQ:
 	kx_typ=NK_IND;
-	if (read(r, &skey, userKeySize)!=userKeySize) {
+	if (read(r, &LM->skey, userKeySize)!=userKeySize) {
 	    logerr(LOG_ERR, "kxchg: read(r)");
 	    return -1;
 	}
-	memcpy(kx_nkind_key, skey, userKeySize);
+	memcpy(kx_nkind_key, LM->skey, userKeySize);
 #ifdef VER_CRC32
-	kx_nkind_crc=(scrc=htonl(crc32(skey, userKeySize)));
+	LM->scrc=htonl(crc32(LM->skey, userKeySize));
 #else
-	kx_nkind_crc=(scrc=block_crc(skey, userKeySize)); /*#*/
+	LM->scrc=block_crc(LM->skey, userKeySize)); /*#*/
 #endif
-	dprintf(("KX: [NK_REQ] sending NK_IND " CRCFORM "\n", scrc));
-	x=send(f, &kx, sizeof(kx), 0); e=errno;
+        put_kx_nkind_crc(&LM->scrc);
+	dprintf(("KX: [NK_REQ] sending NK_IND " CRCFORM "\n", LM->scrc));
+	x=send(f, &LM->kxbuf, sizeof(LM->kxbuf), 0); e=errno;
 	{
 	    struct siocsifcipkey sk;
 	    sk.which=KEY_INVAL+KEY_SEND;
@@ -809,50 +829,55 @@ int kxchg(int f, int r)
 	++ks.req;
 	break;
     case NK_IND:
+        get_kx_nkind_crc(&c);
 	if (
 #ifdef VER_CRC32
-	    crc32(kx_nkind_key, userKeySize)==ntohl(kx_nkind_crc)
+	    crc32(kx_nkind_key, userKeySize)==ntohl(c)
 #else
-	    block_crc(kx_nkind_key, userKeySize)==(kx_nkind_crc) /*#*/
+	    block_crc(kx_nkind_key, userKeySize)==c /*#*/
 #endif
 	    ) {
-	    struct siocsifcipkey sk;
-	    if (dupcheck(kx_nkind_crc)) {
+	    if (dupcheck(c)) {
 		dprintf(("KX: [NK_IND] duplicate\n"));
 		++ks.indb;
 	    } else {
-		sk.which=KEY_RECV;
-		memcpy(&sk.thekey, kx_nkind_key, userKeySize);
-		if (ioctl_setkey(f, device, &sk)<0)
+		LM->sio.which=KEY_RECV;
+		memcpy(&LM->sio.thekey, kx_nkind_key, userKeySize);
+		if (ioctl_setkey(f, device, &LM->sio)<0)
 		    perror("setkey");
 		else {
 		    kx_typ=NK_ACK;
-		    kx_nkack_crc=kx_nkind_crc; /* XX */
+                    put_kx_nkack_crc(&c); /* XX */
 		    dprintf(("KX: [NK_IND] sending NK_ACK " CRCFORM "\n",
-			     (kx_nkack_crc))); /*#*/
-		    x=send(f, &kx, sizeof(kx), 0); e=errno;
+			     c)); /*#*/
+		    x=send(f, &LM->kxbuf, sizeof(LM->kxbuf), 0); e=errno;
 		}
+		memset(&LM->sio, 0, sizeof(LM->sio));
 		++ks.ind;
 	    }
-	    memset(&sk, 0, sizeof(sk));
 	} else {
 	    dprintf(("KX: [NK_IND] invalid\n"));
 	    ++ks.indb;
 	}
 	break;
     case NK_ACK:
-	if (scrc==(kx_nkack_crc)) { /*#*/
-	    struct siocsifcipkey sk;
-	    sk.which=KEY_SEND;
-	    memcpy(&sk.thekey, &skey, userKeySize);
-	    dprintf(("KX: [NK_ACK] got " CRCFORM "\n", scrc));
-	    if (ioctl_setkey(f, device, &sk)<0)
-		perror("setkey");
-	    ++ks.ack;
-	    memset(&sk, 0, sizeof(sk));
-	    memset(&skey, 0, sizeof(skey));
+        get_kx_nkack_crc(&c);
+	if (LM->scrc==c) { /*#*/
+	    if (dupcheck(c)) {
+		dprintf(("KX: [NK_ACK] duplicate\n"));
+		++ks.ackb;
+	    } else {
+		LM->sio.which=KEY_SEND;
+		memcpy(&LM->sio.thekey, &LM->skey, userKeySize);
+		dprintf(("KX: [NK_ACK] got " CRCFORM "\n", LM->scrc));
+		if (ioctl_setkey(f, device, &LM->sio)<0)
+		    perror("setkey");
+		++ks.ack;
+		memset(&LM->sio, 0, sizeof(LM->sio));
+		memset(&LM->skey, 0, sizeof(LM->skey));
+	    }
 	} else {
-	    dprintf(("KX: [NK_ACK] unknown " CRCFORM "\n", scrc));
+	    dprintf(("KX: [NK_ACK] unknown " CRCFORM "\n", c));
 	    ++ks.ackb;
 	}
 	break;
@@ -860,7 +885,7 @@ int kxchg(int f, int r)
 	dprintf(("KX: invalid type %02X\n", kx_typ));
 	++ks.bogus;
     }
-    memset(kx, 0, sizeof(kx));
+    memset(LM->kxbuf, 0, sizeof(LM->kxbuf));
     if (x<0) {
 	errno=e;
 	logerr(LOG_NOTICE, "kxchg: send");
@@ -876,6 +901,8 @@ int kxchg(int f, int r)
 void pspawn(char *s)
 {
     int p, c;
+    if (secchk(s, 0022, 0022, 0))
+        return;
     p=getpid();
     if ((c=fork())<0) {
 	logerr(LOG_ERR, "pspawn: fork");
@@ -892,9 +919,11 @@ void pspawn(char *s)
 	sprintf(bufb, "%d", p);
 	sprintf(bufc, "%s", inet_ntoa(OAaddr(ipaddr)));
 	sprintf(bufd, "%s", inet_ntoa(OAaddr(ptpaddr)));
-        /* put all options but the first (key) into environment */
+        /* put all non-secret options into environment */
         for (o=opts+1; o->oname; ++o)
-            putenv(strdup(optstr(o)));
+            if (o->otyp!=Tsecret)
+                putenv(strdup(optstr(o)));
+        /* should closelog(); ? */
 	close(0); close(1); close(2);
 	(void) open("/dev/null", O_RDWR);
 	(void) dup(0); (void) dup(0);
@@ -908,8 +937,9 @@ void pspawn(char *s)
 
 int main(int argc, char *argv[])
 {
-    int i, ur, fd, sd=-1, pi[2];
+    int i=1, ur, fd, sd=-1;
     time_t t;
+    pid_t pa=0;
     char *st;
     char *pn=strrchr(argv[0], '/');
     if (pn)
@@ -920,55 +950,73 @@ int main(int argc, char *argv[])
     if (getuid() || geteuid()) {
 	fprintf(stderr, "This program must be run by root.\n");
 	exit(1);
-	/* DO NOT remove this check */
+        /* This program is not setuid-safe. If run by local users with
+           arbitrary options, it could easily be tricked into executing
+           any code as root. */
     }
+    if (!(LM=malloc(sizeof(struct lockmem)))) {
+        perror("malloc");
+        exit(1);
+    }
+    memset(LM, 0, sizeof(struct lockmem));
+    xmlock(LM, sizeof(struct lockmem));
+
     /* Snarf options from: standard options file, user-supplied file and
        finally, command line */
     setopt_file(CTLDIR "/options", 0);
-    if ((argc>1) && (!strcmp(argv[1], "-o"))) {
-	if (argc>2) {
-	    setopt_file(argv[2], 1);
-	    argc-=2; argv+=2;
-	} else {
-	    usage(pn);
-	}
+    while ((i=getopt(argc, argv, "io:"))!=EOF) {
+        switch(i) {
+        case 'i': inetd=1; break;
+        case 'o': setopt_file(optarg, 1); break;
+        default: usage(pn);
+        }
     }
-    setopt_cmdline(--argc, ++argv);
+    setopt_cmdline(argc-optind, argv+optind);
     if (!OI(maxerr))
 	OI(maxerr)=8; /* compatibility */
 
     for (i=getdtablesize()-1; i>2; --i)
 	(void) close(i);
-    if (!OI(debug)) {
+
+    if (OI(debug)) {
+	printf("CIPE daemon vers %s (c) Olaf Titz 1996-2000\n",
+	       daemon_version);
+	dumpopt();
+    } else if (!inetd) {
 	/* daemonize */
-        if (pipe(pi)<0) {
-            perror("pipe"); exit(1);
-        }
 	if ((i=fork())<0) {
 	    perror("fork"); exit(1);
 	}
 	if (i) {
             /* wait until the child is ready */
-            close(pi[1]);
-            (void) read(pi[0], &i, 1);
+            setsig(SIGUSR1, sighand);
+            setsig(SIGCHLD, sighand);
+            gotsig=0;
+            while (gotsig!=SIGUSR1) {
+                pause();
+                if (gotsig==SIGCHLD) {
+                    int s;
+                    wait(&s);
+                    exit(WIFEXITED(s)?WEXITSTATUS(s):255);
+                }
+            }
 	    exit(0);
         }
-        close(pi[0]);
-	openlog(pn, LOG_PID, LOGFAC);
+        /* re-lock after fork */
+        xmlock(LM, sizeof(struct lockmem));
+        pa=getppid();
+	openlog(pn, LOG_CONS|LOG_NDELAY|LOG_PID, LOGFAC);
 	/* don't let the syslog socket get in the way of the standard fds */
 	close(0); close(1); close(2);
 	setsid();
-	syslog(LOG_INFO, "CIPE daemon vers %s (c) Olaf Titz 1996-1998",
+	syslog(LOG_INFO, "CIPE daemon vers %s (c) Olaf Titz 1996-2000",
 	       daemon_version);
-    } else {
-	printf("CIPE daemon vers %s (c) Olaf Titz 1996-1998\n",
-	       daemon_version);
-	dumpopt();
     }
 
     setsig(SIGHUP, sighand);
     setsig(SIGINT, sighand);
     setsig(SIGTERM, sighand);
+    setsig(SIGPIPE, sighand);
     errcnt=OI(maxerr);
     if ((ur=open("/dev/urandom", O_RDONLY))<0) {
 	logerr(LOG_ERR, "open(/dev/urandom)"); exit(1);
@@ -995,7 +1043,7 @@ int main(int argc, char *argv[])
     }
     if (i<MAXINT) {
 	struct itimerval t;
-	setsig(SIGALRM, sigignore);
+	setsig(SIGALRM, sighandnot);
 	t.it_interval.tv_sec = i;
 	t.it_interval.tv_usec = 0;
 	t.it_value.tv_sec = i+10;
@@ -1003,10 +1051,11 @@ int main(int argc, char *argv[])
 	setitimer(ITIMER_REAL, &t, 0);
     }
 
-    if (!OI(debug))
-        close(pi[1]);
-    /* sendctl(fd, "\x71starting up", 13); */
+    if (pa>1)
+        kill(pa, SIGUSR1);
+    /* sendctl(fd, "\x71starting up", 13, 1); */
 
+    gotsig=0;
     /* Main loop */
     while (kxchg(fd, ur)>=0) {
         if (gotsig) {
@@ -1025,15 +1074,15 @@ int main(int argc, char *argv[])
         }
         if (t>=pingtime) {
             char buf=CT_PING;
-            sendctl(fd, &buf, 1);
+            sendctl(fd, &buf, 1, 1);
             pingtime=t+OI(ping);
             pongtime=t+OI(toping);
         }
     }
-    sendctl(fd, "\x71going down", 12);
+    gotsig=-1;
+    sendctl(fd, "\x71going down", 12, 0);
     if ((st=getstats()))
         logs(LOG_INFO, "Interface stats %s\n", st);
-    closedev(fd);
     if (OI(debug))
 	printf("KX stats: rreq=%d, req=%d, ind=%d, indb=%d, ack=%d,"
 	       " ackb=%d, unknown=%d\n", ks.rreq, ks.req, ks.ind, ks.indb,
@@ -1043,6 +1092,7 @@ int main(int argc, char *argv[])
 	       " ackb=%d, unknown=%d\n", ks.rreq, ks.req, ks.ind, ks.indb,
 	       ks.ack, ks.ackb, ks.bogus);
     pspawn(OS(ipdown) ? OS(ipdown) : CTLDIR "/ip-down");
+    closedev(fd);
     logs(LOG_INFO, "%s: daemon exiting", device);
     return 0;
 }
