@@ -10,12 +10,11 @@
    as published by the Free Software Foundation; either version
    2 of the License, or (at your option) any later version.
 */
-/* $Id: sock.c,v 1.18.2.2 1999/04/29 09:32:43 olaf Exp $ */
+/* $Id: sock.c,v 1.27 2000/07/29 00:33:35 olaf Exp $ */
 
 #include "cipe.h"
 
 #include <linux/config.h>
-#include <linux/if.h>
 #include <linux/sched.h>
 #include <net/sock.h>
 #include <net/ip.h>
@@ -68,11 +67,13 @@ void cipe_sock_close(struct sock *sock, timeout_t timeout)
     /* Put back the old protocol block and let it do the close */
     sock->prot=c->udp_prot;
     sock->prot->close(sock, timeout);
-    if (c->dev->flags&IFF_UP)
+    if (c->dev->flags&IFF_UP) {
+	rtnl_LOCK();
 	dev_close(c->dev);
-    else
+	rtnl_UNLOCK();
+    } else {
 	cipe_close(c);
-    cipe_unuse_module();
+    }
 }
 
 /* Anything we send to this socket is in fact a key exchange block.
@@ -106,7 +107,7 @@ int cipe_sendmsg(struct sock *sock, struct msghdr *msg, int len
              +blockSize
 #endif
                        ]>=CT_DUMMY) {
-        if (!c->havekey)
+        if (!c->flags&CIPF_HAVE_KEY)
             return -EINVAL;
         buf[KEYXCHGTSPOS-1]='\0';
     } else {
@@ -123,7 +124,7 @@ int cipe_sendmsg(struct sock *sock, struct msghdr *msg, int len
 
     if (c->sockshost) {
 	/* Prepend a socks header. */
-	memset(&sh, 0, 4);
+	memset(&sh, 0, sizeof(sh));
 	sh.atyp=1;
 	sh.dstaddr=c->sockshost;
 	sh.dstport=c->socksport;
@@ -162,47 +163,42 @@ int cipe_sendmsg(struct sock *sock, struct msghdr *msg, int len
     return e;
 }
 
-/* Check if we have a new peer (socks relayer changed, etc) */
+/* Check if we have a new peer */
 static void checkpeer(struct cipe *c, __u32 saddr, __u16 sport)
 {
-    if (c->peeraddr!=saddr || c->peerport!=sport) {
-	unsigned char *x=(unsigned char *)&saddr;
+    if (c->sockshost) {
+	if (c->sockshost==saddr && c->socksport==sport)
+	    return;
+	/* sockshost and socksport contain the real peer's address
+	   and the configured/guessed peer is really the socks relayer! */
+	c->sockshost=saddr;
+	c->socksport=sport;
+    } else {
+	if (c->peeraddr==saddr && c->peerport==sport)
+	    return;
 	c->peeraddr=saddr;
 	c->peerport=sport;
-	printk(KERN_NOTICE "%s: new peer %d.%d.%d.%d:%d\n",
-	       c->dev->name, x[0], x[1], x[2], x[3], ntohs(sport));
     }
+    printk(KERN_NOTICE "%s: new peer %s:%d\n",
+	   c->dev->name, cipe_ntoa(saddr), ntohs(sport));
 }
 
 /* Decrypt a received packet. Requeue it or return kxc block. */
-/* On entry the packet starts with the original IP header (?),
+/* On entry the packet starts with the original UDP header,
    ip_hdr and h.uh are set to the IP and UDP headers. */
 struct sk_buff *cipe_decrypt_skb(struct cipe *c, struct sock *sk,
 				 struct sk_buff *skb, int *offset,
                                  int *msgflag)
 {
-    struct sk_buff *n;
+    struct sk_buff *n=NULL;
     int length;
     __u32 rsaddr;
     __u16 rsport;
 
 #ifdef DEBUG
     if (cipe_debug&DEB_PKIN)
-        cipe_dump_packet("received", skb);
+        cipe_dump_packet("received", skb, 1);
 #endif
-    n=alloc_skb(skb->len, GFP_KERNEL);
-    if (!n) {
-	printk(KERN_WARNING "%s: cipe_decrypt_skb(): out of memory\n",
-	       c->dev->name);
-	++c->stat.rx_dropped;
-	return NULL;
-    }
-    n->sk=NULL;
-#ifndef LINUX_21
-    n->free=1;
-#endif
-    n->dev=c->dev;
-
     length=ntohs(skb->h.uh->len)-sizeof(struct udphdr);
 #if 0 /* UDP should check this */
     if (length!=skb->len-sizeof(struct udphdr)) {
@@ -213,12 +209,28 @@ struct sk_buff *cipe_decrypt_skb(struct cipe *c, struct sock *sk,
     }
 #endif
 
+    n=alloc_skb(skb->len, GFP_KERNEL);
+    if (!n) {
+	printk(KERN_WARNING "%s: cipe_decrypt_skb: out of memory\n",
+	       c->dev->name);
+	++c->stat.rx_dropped;
+	return NULL;
+    }
+#if 0 /*def LINUX_21*/
+    if (skb->sk)
+	skb_set_owner_r(n, skb->sk);
+#endif
+#ifndef LINUX_21
+    n->free=1;
+#endif
+    n->dev=c->dev;
+
     /* Copy the datagram into new buffer, skip IP header */
     /* We must copy here because raw sockets get only a clone of the
        skbuff, so they would receive the plaintext */
     dprintk(DEB_INP, (KERN_INFO DEVNAME
-                      ": sa=%08lX da=%08lX us=%d ud=%d ul=%d\n",
-                      ntohl(saddr(skb)), ntohl(daddr(skb)),
+                      ": sa=%s da=%s us=%d ud=%d ul=%d\n",
+                      cipe_ntoa(saddr(skb)), cipe_ntoa(daddr(skb)),
                       ntohs(skb->h.uh->source), ntohs(skb->h.uh->dest),
                       ntohs(skb->h.uh->len)));
 
@@ -231,21 +243,15 @@ struct sk_buff *cipe_decrypt_skb(struct cipe *c, struct sock *sk,
     rsport=skb->h.uh->source;
     skb_put(n,skb->len);
     memcpy(n->data, skb->h.raw, skb->len);
-    /*n->ip_hdr=NULL;*/
     n->h.uh=(struct udphdr *)n->data;
-#if 0
-    dprintk4(DEB_INP, KERN_INFO, "UDP: s=%d d=%d l=%d c=%04X\n",
-	     ntohs(n->h.uh->source), ntohs(n->h.uh->dest),
-	     ntohs(n->h.uh->len), n->h.uh->check);
-#endif
 
     if (c->sockshost) {
 	/* Pull out the SOCKS header and correct the peer's address. */
 	struct sockshdr *sh;
 	sh=(struct sockshdr *)skb_pull(n, sizeof(struct udphdr));
-	dprintk(DEB_INP, (KERN_INFO "socks: fr=%d at=%d da=%08lX dp=%d\n",
+	dprintk(DEB_INP, (KERN_INFO "socks: fr=%d at=%d da=%s dp=%d\n",
                           sh->frag, sh->atyp,
-                          ntohl(sh->dstaddr), ntohs(sh->dstport)));
+                          cipe_ntoa(sh->dstaddr), ntohs(sh->dstport)));
 	if (sh->frag || (sh->atyp!=1))
 	    goto error;
 	/* Pull out UDP header */
@@ -255,7 +261,8 @@ struct sk_buff *cipe_decrypt_skb(struct cipe *c, struct sock *sk,
 	n->ip_hdr=(struct iphdr *)skb_pull(n, sizeof(struct sockshdr));
 #endif
 	/***saddr(n)=sh->dstaddr;*/
-	n->h.uh->source=sh->dstport;
+        rsaddr=sh->dstaddr;
+	rsport=n->h.uh->source=sh->dstport;
 	length-=sizeof(struct sockshdr);
 	*offset=sizeof(struct sockshdr)+sizeof(struct udphdr);
     } else {
@@ -269,10 +276,21 @@ struct sk_buff *cipe_decrypt_skb(struct cipe *c, struct sock *sk,
 	n->h.uh->source=rsport;
 	*offset=sizeof(struct udphdr);
     }
+    n->mac.raw=n->data; /* no hardware header */
 
-    if (c->havekey) {
+    if (c->flags&CIPF_HAVE_KEY) {
+	if (length%blockSize) {
+	    printk(KERN_INFO "%s: got bogus length=%d\n", c->dev->name,
+		   length);
+	    goto error;
+	}
         switch (cipe_decrypt(c, n->data, &length)) {
 	case TW_DATA:
+            if (!c->flags&CIPF_MAY_STKEY && !c->flags&CIPF_HAVE_RKEY) {
+                printk(KERN_ERR "%s: got data using invalid static key\n",
+                       c->dev->name);
+                goto error;
+            }
 	    break;
 	case TW_CTRL:
             /* return it as a control block - out of band datagram */
@@ -280,6 +298,10 @@ struct sk_buff *cipe_decrypt_skb(struct cipe *c, struct sock *sk,
             /* fall thru */
 	case TW_NEWKEY:
 	    /* return it as key exchange block - proper UDP datagram */
+	    dprintk(DEB_INP, (KERN_DEBUG "TW_NEWKEY data=%p len=%d length=%d\n", n->data, n->len, length));
+#ifdef LINUX_21
+	    get_fast_time(&n->stamp);
+#endif
 	    skb_trim(n, length);
 	    checkpeer(c, rsaddr, rsport);
 #if 0
@@ -293,10 +315,11 @@ struct sk_buff *cipe_decrypt_skb(struct cipe *c, struct sock *sk,
 	    ++c->stat.rx_crc_errors;
 	    goto error;
 	}
-    } else if (!c->mayclear) {
+    } else if (!c->flags&CIPF_MAY_CLEAR) {
 	goto error;
     }
 
+    dprintk(DEB_INP, (KERN_DEBUG "TW_DATA data=%p len=%d length=%d\n", n->data, n->len, length));
     skb_trim(n, length);
     checkpeer(c, rsaddr, rsport);
     /* adjust pointers */
@@ -308,27 +331,22 @@ struct sk_buff *cipe_decrypt_skb(struct cipe *c, struct sock *sk,
     memset(n->proto_priv, 0, sizeof(struct options));
 #endif
 
-    n->mac.raw=n->data; /* no hardware header */
-    n->protocol = htons(ETH_P_IP);
+    n->protocol = __constant_htons(ETH_P_IP);
     n->ip_summed = 0;
-#ifdef DEBUG
-    if (cipe_debug&DEB_INP) {
-#ifdef LINUX_21
-        unsigned char *x=(unsigned char *)&(saddr(n));
-#else
-        unsigned char *x=n->data;
-#endif
-	printk(KERN_INFO
-	       "%s: real src %d.%d.%d.%d dst %d.%d.%d.%d len %d\n",
-	       n->dev->name,
-	       x[0], x[1], x[2], x[3], x[4], x[5], x[6], x[7], length);
-    }
-#endif
+    n->pkt_type = PACKET_HOST;
+    dprintk(DEB_INP, (KERN_INFO "%s: real src %s dst %s len %d\n",
+                      n->dev->name, cipe_ntoa(saddr(n)),
+                      cipe_ntoa(daddr(n)), length));
 
 #ifdef DEBUG
     if (cipe_debug&DEB_PKIN)
-        cipe_dump_packet("decrypted", n);
+        cipe_dump_packet("decrypted", n, 1);
+#ifdef LINUX_23
+    dprintk(DEB_INP, (KERN_DEBUG "%s state=%ld refcnt=%d\n", n->dev->name, n->dev->state,
+		      atomic_read(&n->dev->refcnt)));
 #endif
+#endif
+
     /* requeue */
     netif_rx(n);
 #ifdef LINUX_21
@@ -343,8 +361,8 @@ struct sk_buff *cipe_decrypt_skb(struct cipe *c, struct sock *sk,
 #endif
  error:
     ++c->stat.rx_errors;
-    /*n->sk=NULL;*/
-    kfreeskb(n, FREE_READ);
+    if (n)
+	kfreeskb(n, FREE_READ);
     return NULL;
 }
 
@@ -396,13 +414,13 @@ int cipe_recvmsg(struct sock *sk, struct msghdr *msg, int len,
 		break;
 	    }
 	    skn=cipe_decrypt_skb(c, sk, skb, &of, &(msg->msg_flags));
-            dprintk(DEB_INP, (KERN_INFO "%s: returning %d flags %d\n",
-                              c->dev->name, skb->len, msg->msg_flags));
 	    skb_free_datagram(sk, skb);
 	    skb=skn;
 	} while (skb==NULL);
 
-  	copied = skb->len;
+	dprintk(DEB_INP, (KERN_INFO "%s: returning %d flags %04x\n",
+			  c->dev->name, skb->len, msg->msg_flags));
+  	copied = skb->len-of;
 	if (copied > len)
 	{
 		copied = len;
@@ -410,6 +428,10 @@ int cipe_recvmsg(struct sock *sk, struct msghdr *msg, int len,
 		msg->msg_flags |= MSG_TRUNC;
 #endif
 	}
+	if (copied<=0) {
+	    printk(KERN_ERR "cipe_recvmsg: bogus length %d\n", copied);
+	    goto out_free;
+	}
 
   	/*
   	 *	FIXME : should use udp header size info value
@@ -432,30 +454,10 @@ int cipe_recvmsg(struct sock *sk, struct msghdr *msg, int len,
 		sin->sin_family = AF_INET;
 		sin->sin_port = skb->h.uh->source;
                 sin->sin_addr.s_addr = daddr(skb);
-#ifdef CONFIG_IP_TRANSPARENT_PROXY
-		/* Does this ever apply here? */
-		if (flags&MSG_PROXY)
-		{
-			/*
-			 * We map the first 8 bytes of a second sockaddr_in
-			 * into the last 8 (unused) bytes of a sockaddr_in.
-			 * This _is_ ugly, but it's the only way to do it
-			 * easily,  without adding system calls.
-			 */
-			struct sockaddr_in *sinto =
-				(struct sockaddr_in *) sin->sin_zero;
-
-			sinto->sin_family = AF_INET;
-			sinto->sin_port = skb->h.uh->dest;
-			sinto->sin_addr.s_addr = saddr(skb);
-		}
-#endif
 	}
         er=copied;
 
-#ifdef LINUX_21
  out_free:
-#endif
 	if (skb==skn)
 	    /* the buffer was a copy made by decryption */
 	    kfreeskb(skb, FREE_READ);
@@ -469,12 +471,15 @@ int cipe_recvmsg(struct sock *sk, struct msghdr *msg, int len,
 
 #include <linux/file.h>
 
-int cipe_attach(struct device *dev, struct siocsifcipatt *parm)
+int cipe_attach(struct NET_DEVICE *dev, struct siocsifcipatt *parm)
 {
     struct file *file;
     struct inode *inode;
     struct socket *sock;
     struct sock *sk;
+#if 0
+    struct in_device *indev;
+#endif
     struct cipe *c0;
     DEVTOCIPE(dev,c,-ENODEV);
 
@@ -506,10 +511,13 @@ int cipe_attach(struct device *dev, struct siocsifcipatt *parm)
 	return(-EBUSY); /* socket is already attached */
 
     cipe_use_module();
+    c->owner=current->pid;
     c->sock=sk;
     c->peeraddr=sk->daddr;
     c->peerport=sk->dport;
     c->myaddr=sk->saddr;
+    if (c->flags&CIPF_MAY_DYNIP)
+        sk->rcv_saddr=0;
     c->myport=htons(sk->num);
     /* Disconnect socket, we might receive from somewhere else */
     sk->daddr=0;
@@ -528,6 +536,25 @@ int cipe_attach(struct device *dev, struct siocsifcipatt *parm)
     c->cipe_proto.recvmsg=cipe_recvmsg;
     sk->prot=&c->cipe_proto;
 
+#if 0
+    /* (Try to) Set our dummy hardware address from the IP address. */
+    /* This does not work, because the IP address is set
+       _after_ the attach... */
+    if ((indev=dev->ip_ptr)) {
+	struct in_ifaddr **ifap = NULL;
+	struct in_ifaddr *ifa = NULL;
+	for (ifap=&indev->ifa_list; (ifa=*ifap) != NULL; ifap=&ifa->ifa_next)
+	    if (strcmp(dev->name, ifa->ifa_label) == 0)
+		break;
+	if (ifa) {
+	    char *x=(char *)&ifa->ifa_local;
+	    dev->dev_addr[3]=x[1]|0x80;
+	    dev->dev_addr[4]=x[2];
+	    dev->dev_addr[5]=x[3];
+	}
+    }
+#endif
+
     return(0);
 }
 
@@ -536,7 +563,7 @@ int cipe_attach(struct device *dev, struct siocsifcipatt *parm)
 
 #define sreturn(x) {sti(); return((x));}
 
-int cipe_attach(struct device *dev, struct siocsifcipatt *parm)
+int cipe_attach(struct NET_DEVICE *dev, struct siocsifcipatt *parm)
 {
     struct file *file;
     struct inode *inode;
@@ -568,10 +595,13 @@ int cipe_attach(struct device *dev, struct siocsifcipatt *parm)
 	sreturn(-EBUSY); /* socket is already attached */
 
     cipe_use_module();
+    c->owner=current->pid;
     c->sock=sk;
     c->peeraddr=sk->daddr;
     c->peerport=sk->dummy_th.dest;
     c->myaddr=sk->saddr;
+    if (c->flags&CIPF_MAY_DYNIP)
+        sk->rcv_saddr=0;
     c->myport=sk->dummy_th.source;
     /* Disconnect socket, we might receive from somewhere else */
     sk->daddr=0;
@@ -588,7 +618,7 @@ int cipe_attach(struct device *dev, struct siocsifcipatt *parm)
     /* Fill an otherwise unused field in the sock struct with this info.
        Actually, this is very similar to a packet socket!
        The ugly cast saves us one deref in the actual ops */
-    sk->protinfo.af_packet.bound_dev=(struct device *)c;
+    sk->protinfo.af_packet.bound_dev=(struct NET_DEVICE *)c;
     sk->no_check=1; /* our packets are checksummed internally */
 
     sti();
@@ -607,7 +637,7 @@ void cipe_fakenkey(struct cipe *c, char typ)
     struct sk_buff *skb=alloc_skb(len, GFP_ATOMIC);
 
     if (!skb) {
-        printk(KERN_WARNING "%s: cipe_fakenkey(): out of memory\n",
+        printk(KERN_WARNING "%s: cipe_fakenkey: out of memory\n",
                c->dev->name);
         return; /* not much we can do */
     }
@@ -628,12 +658,10 @@ void cipe_fakenkey(struct cipe *c, char typ)
     skb->free=1; /* Discard after use */
     skb->ip_hdr=NULL;
 #endif
-    skb->mac.raw=NULL;
-
     skb->h.uh->source=0; /* mark as KXC packet */
     skb->h.uh->dest=c->myport;
     len-=sizeof(struct udphdr);
-    skb->h.uh->len=len;
+    skb->h.uh->len=htons(len);
     skb->h.uh->check=0;
     skb->mac.raw=skb->data; /* no hardware header */
 
@@ -643,7 +671,7 @@ void cipe_fakenkey(struct cipe *c, char typ)
     skb->data[sizeof(struct udphdr)]=typ;
 
     if (sock_queue_rcv_skb(c->sock, skb)<0) {
-        printk(KERN_WARNING "%s: cipe_fakenkey(): enqueuing failed\n",
+        printk(KERN_WARNING "%s: cipe_fakenkey: enqueuing failed\n",
                c->dev->name);
         kfreeskb(skb, FREE_WRITE);
     }
